{
  Susie Plugin Helper Unit
  2002/11/10-
}

unit SPIUtil;

interface

uses Windows, spi_api, simplelist, diblib;

var
  PluginList: TSimpleList;
const
  GF_UNKNOWN = 0;
  GF_IMAGE = 1;
  GF_ARCHIVE = 2;
type
  PFileList = ^TFileList;
  TFileList = record
    FileName: array[0..MAX_PATH] of Char;
    FileMTime: TFileTime;
    IsImage: Boolean;
    FileInfo: TFileInfo; // for 00AM plugin
  end;

function GetPluginPath: string;
function GetFormat(const FileName: string; Buf: PChar; var PluginHandle: THandle): Integer;
function ExtractImage(PluginHandle: THandle; pSrc: PChar; SrcSize: Integer;
  BitmapInfo: PBitmapInfo; var pDst: Pointer; var DstSize: Integer;
  var ColorDepth: Integer; var TextInfo: string; var HBm: HLOCAL): Integer;
function GetPluginName(PluginHandle: THandle): string;
procedure GetFileListFromArchive(h: THandle; const FileName: string; var FileList: TSimpleList);
procedure GetFileFromArchive(const FileName: string; var pBuf: PChar;
  var size: Integer; h: THandle; const FileList: PFileList);

implementation

//
// Susie plug-in ̏ꏊ擾
//
function GetPluginPath: string;
var
  data: array[0..MAX_PATH] of Char;
  size: DWORD;
  hkResult: HKEY;
begin
  FillChar(data, SizeOf(data), 0);
  if RegOpenKeyEx(HKEY_CURRENT_USER, 'Software\Takechin\Susie\Plug-in', 0,
    KEY_QUERY_VALUE, hkResult) = ERROR_SUCCESS then
  begin
    size := SizeOf(data);
    RegQueryValueEx(hkResult, 'Path', nil, nil, @data, @size);
    RegCloseKey(hkResult);
  end;
  Result := data;
end;

function FileListCompare(Item1, Item2: Pointer): Integer;
begin
  Result := lstrcmp(PFileList(Item1).FileName, PFileList(Item2).FileName);
end;

function time_tToFileTime(t: Longint): TFileTime;
var
  st: TSystemTime;
  ft: TFileTime;
begin
  st.wYear := 1970;
  st.wMonth := 1;
  st.wDay := 1;
  st.wHour := 0;
  st.wMinute := 0;
  st.wSecond := 0;
  st.wMilliseconds := 0;
  SystemTimeToFileTime(st, ft);
  Comp(Result) := Comp(ft) + t * 10000000.0;
end;

procedure GetFileListFromArchive(h: THandle; const FileName: string; var FileList: TSimpleList);
var
  p: PChar;
  GetArchiveInfo: TGetArchiveInfo;
  hinf: HLOCAL;
  pItem: PFileList;
  s: string;
begin
  GetArchiveInfo := GetProcAddress(h, 'GetArchiveInfo');
  GetArchiveInfo(PChar(FileName), 0, 0, hinf);
  p := LocalLock(hinf);
  if p = nil then Exit;
  while p^ <> #0 do
  begin
    if PFileInfo(p).FileSize > 0 then // FileSize = 0 ̓fBNg
    begin
      New(pItem);
      s := PFileInfo(p).Path + PFileInfo(p).FileName;
      lstrcpyn(pItem.FileName, PChar(s), SizeOf(pItem.FileName));
      pItem.FileInfo := PFileInfo(p)^;
      pItem.FileMTime := time_tToFileTime(pItem.FileInfo.TimeStamp);
      pItem.IsImage := True;
      FileList.Add(pItem);
    end;
    Inc(p, SizeOf(TFileInfo));
  end;
  FileList.Sort(FileListCompare);
end;

procedure GetFileFromArchive(const FileName: string; var pBuf: PChar;
  var size: Integer; h: THandle; const FileList: PFileList);
var
  GetFile: TGetFile;
  HDst: HLOCAL;
  p: PChar;
begin
  GetFile := GetProcAddress(h, 'GetFile');
  GetFile(PChar(FileName), FileList.FileInfo.Position, @HDst, $100, nil, 0);
  p := LocalLock(HDst);
  size := LocalSize(HDst);
  GetMem(pBuf, size);
  Move(p^, pBuf^, size);
  LocalUnLock(HDst);
  LocalFree(HDst);
end;

function ProgressCallback(nNum, nDemon: Integer; lData : Longint): Integer; stdcall;
begin
  Result := 0;
end;

function GetPluginName(PluginHandle: THandle): string;
var
  GetPluginInfo: TGetPluginInfo;
  buf: array[0..1024] of Char;
begin
  GetPluginInfo := GetProcAddress(PluginHandle, 'GetPluginInfo');
  GetPluginInfo(1, @buf, SizeOf(buf));
  Result := buf;
end;

function ExtractImage(PluginHandle: THandle; pSrc: PChar; SrcSize: Integer;
  BitmapInfo: PBitmapInfo; var pDst: Pointer; var DstSize: Integer;
  var ColorDepth: Integer; var TextInfo: string; var HBm: HLOCAL): Integer;
var
  GetPicture: TGetPicture;
  GetPictureInfo: TGetPictureInfo;
  PicInfo: TPictureInfo;
  HBInfo: HLOCAL;
  OffBits: Integer;
  p: PChar;
  pbmi: PBitmapInfo;
begin
  //DelphĩǗ}l[WXy[XfꂽubN
  //ԂꍇSetDIBitɎs\̂GlobalLockg

  TextInfo := '';
  if PluginHandle = 0 then
  begin
    HBm := 0;
    OffBits := PBitmapFileHeader(pSrc).bfOffBits;
    DstSize := SrcSize - OffBits;
    pbmi :=  @pSrc[SizeOf(TBitmapFileHeader)];
    // BITMAPINFOHEADER  biSizeImage sȏꍇC
    if (pbmi.bmiHeader.biSizeImage = 0) and (pbmi.bmiHeader.biCompression <> BI_RGB) then
      pbmi.bmiHeader.biSizeImage := DstSize;
    Move(pbmi^, BitmapInfo^, OffBits - SizeOf(TBitmapFileHeader));

    if BitmapInfo.bmiHeader.biSize = SizeOf(TBitmapCoreHeader) then
      CoreInfoToBitmapInfo(BitmapInfo, PBitmapCoreInfo(@pSrc[SizeOf(TBitmapFileHeader)]));

    // BI_RGB  16bit DIB ̏ꍇ RGB555(=15bit) (Microsoft Knowledge Base J028984)
    if (BitmapInfo.bmiHeader.biBitCount = 16) and (BitmapInfo.bmiHeader.biCompression = BI_RGB) then
      ColorDepth := 15
    else
      ColorDepth := BitmapInfo.bmiHeader.biBitCount;

    // RLEWJ
    if (BitmapInfo.bmiHeader.biCompression = BI_RLE8) or (BitmapInfo.bmiHeader.biCompression = BI_RLE4) then
      DstSize := RLEtoDIB24(BitmapInfo, pDst, PBitmapInfo(@pSrc[SizeOf(TBitmapFileHeader)])^,
        @pSrc[OffBits])
    else begin
      pDst := Pointer(GlobalAlloc(0, DstSize));
      Move(pSrc[OffBits], pDst^, DstSize);
    end;

    Result := 0;
  end
  else begin
    GetPictureInfo := GetProcAddress(PluginHandle, 'GetPictureInfo');
    GetPictureInfo(pSrc, SrcSize, 1, PicInfo);
    ColorDepth := PicInfo.colorDepth;
    if PicInfo.hInfo <> 0 then
    begin
      p := LocalLock(PicInfo.hInfo);
      SetLength(TextInfo, Length(p));
      Move(p^, PChar(TextInfo)^, Length(p));
      LocalUnLock(PicInfo.hInfo);
      LocalFree(PicInfo.hInfo);
    end;

    GetPicture := GetProcAddress(PluginHandle, 'GetPicture');
    // ProgressCallbacknilƗvOĈŃ_~[֐n
    Result := GetPicture(pSrc, SrcSize, 1, @HBInfo, @HBm, @ProgressCallback, 0);
    if Result = 0 then
    begin
      if HBInfo <> 0 then
        Move((LocalLock(HBInfo))^, BitmapInfo^, LocalSize(HBInfo));
      LocalUnLock(HBInfo);
      LocalFree(HBInfo);
      pDst := LocalLock(HBm);
      DstSize := LocalSize(HBm);
    end;
  end;
end;

function GetFormat(const FileName: string; Buf: PChar; var PluginHandle: THandle): Integer;
var
  i: Integer;
  h: THandle;
  PluginVersion: array[0..2047] of Char;
  GetPluginInfo: TGetPluginInfo;
  IsSupported: TIsSupported;
begin
  PluginHandle := 0;

  if (Buf[0] = 'B') and (Buf[1] = 'M') then
  begin
    Result := GF_IMAGE;
    Exit;
  end;

  // SPI
  for i := 0 to PluginList.Count - 1 do
  begin
    h := LoadLibrary(PluginList[i]);
    try
      GetPluginInfo := GetProcAddress(h, 'GetPluginInfo');
      if not Assigned(GetPluginInfo) then Continue;
      GetPluginInfo(0, @PluginVersion, SizeOf(PluginVersion));

      IsSupported := GetProcAddress(h, 'IsSupported');
      if not Assigned(IsSupported) then Continue;
      if IsSupported(PChar(FileName), DWORD(Buf)) <> 0 then
      begin
        PluginHandle := h;
        if PluginVersion = '00IN' then
          Result := GF_IMAGE
        else if PluginVersion = '00AM' then
          Result := GF_ARCHIVE
        else begin
          PluginHandle := 0;
          Result := GF_UNKNOWN;
        end;
        Exit;
      end;
    finally
      if PluginHandle = 0 then FreeLibrary(h);
    end;
  end;

  Result := GF_UNKNOWN;
end;

end.
